unit MMMainFrm;
{------------------------------------------------------------------------------}
(*
  --- DEMO ---
  Mail merging application supporting master-detail relationship and
  the following field values:
  - multiline text;
  - pictures;
  - tables (not used in this demo);
  Fields are any strings inside {}.

  --- SUPPORTED FIELDS ---
  {DATE} - current date and time; result is a plain text.
  {<DATASET>:<DATAFIELD>},
    where <DATASET> - dataset (db table) name, <
    DATAFIELD> - name of field in this db table;
    result is a plain text or picture.
    Example: {Kingdom:Name}

  --- MASTER-DETAIL ---
  A template may contain tables linked to datasets.
  Such tables are recognized by a special format of their item text.
  It has format: <DATASET> <FIRST REPEATED ROW> <COUNT OF REPEATED ROWS>
  Example: Kingdom 1 2
  A group of repeated rows corresponds to one record in the dataset. They are
  repeated as many times as many records in this dataset, taking master-detail
  relationship into account.
  A group of repeated rows may contain merged cells, both vertically and
  horizontally. Obviously, it is not allowed to merge cells across repeated
  and fixed rows.
  If such table is inserted in a repeated row of another such table,
  a master-detail relationship is established. A nested table becomes a detail
  table. The program gets Filter string for detail tables from the master table
  using GetFilterFromMaster procedure.
  Tables can be nested on any depth.

  --- TABLES IN THIS DEMO ---
  This demo shows a fragment of biological classification.

  Kingdom.db contains biological kingdoms.
  Key field: ID.
  This is a master table for Phylum.db and Class.db.

  Phylum.db contains biological phyla.
  Key fields: Kingdom (reference to Kingdom:ID) and ID.
  This is a master table for Class.db.

  Class.db contains divisions of phyla (not necessary biological classes).
  Key fields: Kingdom and Phylum (reference to Kingdom:ID and Phylum:ID) and ID.

  --- MASTER-DETAIL IN THIS DEMO ---
  If GetFilterFromMaster is called for Kingdom.db, it returns:
  Kingdom=<VALUE>
  where <VALUE> is a value of Kingdom:ID for the current
  instance of repeated record. This filter can be applied to tables linked
  with Phylum.db or Class.db inserted in this row.
  If GetFilterFromMaster is called for Phylum.db, it returns:
  Kingdom=<VALUE1> and Phylum=<VALUE2>
  where <VALUE1> is a value of Phylum:Kingdom,
  <VALUE2> is a value of Phylum:ID. This filter can be applied to tables linked
  with Class.db inserted in this row.

  --- CODE STRUCTURE IN THIS DEMO ---
  The code in this demo consists of two parts.
  The first (smaller) part depends on the demo data: opens tables, creates a
    demo template, returns filters for demo tables, processes non-db fields.
  The second part (after the '{======}' line) can be copied to your application
  unchanged.

  --- POSSIBLE MODIFICATIONS ---
  In this demo, there are three 1:N relationships:
    Kingdom.db - Phylum.db
    Phylum.db - Class.db
    Kingdom.db - Class.db
  A record in Phylum.db contains a reference to Kingdom.db, so you can get the
  complete filter from Phylum.db to filter Class.db.

  Imagine another situation: three tables, but only two 1:N relationships:
    A.db - C.db
    B.db - C.db
  Relation A.db - B.db is not necessary 1:N, it may be M:N.
  However, if a table-C is inserted in a table-B,
  and a table-B is inserted in a table-A,
  it may be expected that table-C is filtered both by A and B.
  In this case, you cannot get Filter for table C.db only from table B.db.
  You need to combine both the filters received from A.db and B.db.
  The place in code where you can do it is marked with the comment {!FILTER!}

*)
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, RVScroll, RichView, StdCtrls, RVStyle, RVTable, CRVData, RVItem,
  RVEdit, RVUni, RVTypes, DB, DBTables;

{$I RV_Defs.inc} // required for ConvertStringToItemText()

type


  { type of field value }
  TMMFieldType = (mmftText, mmftPicture, mmftTable, mmftError);

  TForm1 = class(TForm)
    rvOutput: TRichView;
    rvsOutput: TRVStyle;
    rvsTemplate: TRVStyle;
    Label1: TLabel;
    Label2: TLabel;
    rve: TRichViewEdit;
    Button1: TButton;
    tblKingdom: TTable;
    tblPhylum: TTable;
    tblSubphylum: TTable;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
    Filter: String;
    procedure CreateTemplate;
    function GetFieldType(const FieldName: String): TMMFieldType;
    function GetFieldValueFromDatabase(const FieldName: String): String;
    function GetPicture(const FieldName: String): TGraphic;
    function GetTable(RVData: TCustomRVData; const FieldName: String): TRVTableItemInfo;
    function GetFilterFromMaster(DataSet: TDataSet): String;
    procedure FillFields(RVData: TCustomRVData; StartItemNo: Integer);
    procedure InsertField(RVData: TCustomRVData;
      var ItemNo, Pos: Integer; const FieldName: String);
    procedure InsertMultilineText(RVData: TCustomRVData;
      var ItemNo, Pos: Integer; const Value: String);
    procedure InsertItem(RVData: TCustomRVData; var ItemNo, Pos: Integer;
      const Item: TCustomRVItemInfo);
    function GetDataSet(const TableName: String): TDataSet;
    function ParseTableText(s: String; RowCount: Integer;
      var DataSet: TDataSet;
      var RepRowStart, RepRowCount: Integer): Boolean;
    function GetNonDBFieldType(const FieldName: String): TMMFieldType;
    function GetNonDBFieldValue(const FieldName: String): String;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{------------------------------------------------------------------------------}
procedure TForm1.FormCreate(Sender: TObject);
begin
  try
  if not Session.IsAlias('RVBIODEMO') then
    Session.AddStandardAlias('RVBIODEMO', ExtractFilePath(Application.ExeName), '');
  except;
  end;
  tblKingdom.Open;
  tblPhylum.Open;
  tblSubphylum.Open;
  CreateTemplate;
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormDestroy(Sender: TObject);
begin
  Session.DeleteAlias('RVBIODEMO');
end;
{------------------------------------------------------------------------------}
// Clearing all table cells (removing empty text items)
// This procedure is called before adding content to cells
procedure ClearAllCells(table: TRVTableItemInfo);
var r,c: Integer;
begin
  for r := 0 to table.RowCount-1 do
    for c := 0 to table.ColCount-1 do
      if table.Cells[r,c]<>nil then
        table.Cells[r,c].GetRVData.Clear;
end;
{------------------------------------------------------------------------------}
// Creating a sample template. You can edit it
procedure TForm1.CreateTemplate;
var table1, table2, table3: TRVTableItemInfo;
begin
   rve.Clear;
   rve.AddNL('Report created: {DATE}', 0, 0);

   table1 := TRVTableItemInfo.CreateEx(4,1,rve.RVData); // linked to Kingdom.db
   ClearAllCells(table1);
   table1.BorderWidth := 1;
   table1.BorderStyle := rvtbColor;
   table1.Cells[0,0].AddNL('Life', 4, 1);;
   table1.Cells[1,0].Color := clBlack;
   table1.Cells[1,0].AddNL('Kingdom {Kingdom:Name}', 2, 0);;
     table2 := TRVTableItemInfo.CreateEx(2,3,rve.RVData); // linked to Phylum.db
     table2.MergeCells(1, 0, 3, 1, True);
     ClearAllCells(table2);
     table2.Cells[0,0].Color := clSkyBlue;
     table2.Cells[0,0].AddNL('Name: ', 3, 0);
     table2.Cells[0,0].AddNL('{Phylum:Name}', 0, -1);;
     table2.Cells[0,1].Color := clSkyBlue;
     table2.Cells[0,1].AddNL('Meaning: ', 3, 0);;
     table2.Cells[0,1].AddNL('{Phylum:Meaning}', 0, -1);;
     table2.Cells[0,2].Color := clSkyBlue;
     table2.Cells[0,2].AddNL('Common name: ', 3, 0);
     table2.Cells[0,2].AddNL('{Phylum:CommonName}', 0, -1);
       table3 := TRVTableItemInfo.CreateEx(2,2,rve.RVData); // linked to Class.db
       table3.MergeCells(0, 0, 2, 1, True);
       ClearAllCells(table3);
       table3.Cells[0,0].AddNL('{Class:Name}', 3, 1);
       table3.Cells[1,0].AddNL('{Class:Description}', 0, 0);
       table3.Cells[1,1].AddNL('{Class:Picture}', 0, 1);
     table2.Cells[1,0].AddItem('Class 0 2', table3);
   table1.Cells[2,0].AddItem('Phylum 0 2', table2);
   table1.Cells[3,0].AddNL('Data from Wikipedia, the free encyclopedia', 1, 1);
   rve.AddItem('Kingdom 1 2', table1);
   rve.Format;
end;
{------------------------------------------------------------------------------}
// Returns dataset by its name
function TForm1.GetDataSet(const TableName: String): TDataSet;
var s: String;
begin
  s := AnsiLowerCase(TableName);
  if s='class' then
    Result := tblSubphylum
  else if s='phylum' then
    Result := tblPhylum
  else if s='kingdom' then
    Result := tblKingdom
  else
    Result := nil;
end;
{------------------------------------------------------------------------------}
// Returns Filter for detail tables corresponding to the current record of DataSet
function TForm1.GetFilterFromMaster(DataSet: TDataSet): String;
begin
  if (DataSet=tblKingdom) then
    Result := 'Kingdom = '+QuotedStr(DataSet.FieldByName('ID').AsString)
  else if DataSet=tblPhylum then
    Result := 'Kingdom = '''+DataSet.FieldByName('Kingdom').AsString+
      ''' and Phylum = '''+DataSet.FieldByName('ID').AsString+''''
  else
    Result := '';
end;
{------------------------------------------------------------------------------}
// Returns type of field not linked to dataset.
// It must return mmftText for recognized fields, mmftError otherwise.
function TForm1.GetNonDBFieldType(const FieldName: String): TMMFieldType;
begin
  if LowerCase(FieldName)='date' then
    Result := mmftText
  else
    Result := mmftError
end;
{------------------------------------------------------------------------------}
// Returns value of field not linked to dataset.
function TForm1.GetNonDBFieldValue(const FieldName: String): String;
begin
  if LowerCase(FieldName)='date' then
    Result := DateTimeToStr(Now)
  else
    Result := '';
end;
{------------------------------------------------------------------------------}
// (NOT USED IN THIS DEMO
// Creates a table by the field code.
//  This example creates table of the specified size with "Sample Table" text.
//   In your application, if the field code is invalid, you can return nil.
function TForm1.GetTable(RVData: TCustomRVData; const FieldName: String): TRVTableItemInfo;
var Size, i, BestWidth: Integer;
begin
  Size := StrToIntDef(Copy(FieldName, 6, Length(FieldName)), 3);
  Result := TRVTableItemInfo.CreateEx(Size, Size, RVData);
  Result.BorderWidth := 3;
  Result.CellBorderWidth := 1;
  Result.BestWidth := -100;
  BestWidth := 100;
  Result.MergeCells(0, 0, Result.ColCount, 1, True);
  Result.Cells[0,0].Clear();
  Result.Cells[0,0].AddNL('Sample Table', 0, 0);
end;

{------------------------------------------------------------------------------}
// Generating output
procedure TForm1.Button1Click(Sender: TObject);
var Stream: TStream;
begin
  Screen.Cursor := crHourGlass;
  try
    // copying rve to rvOutput
    Stream := TMemoryStream.Create;
    rve.SaveRVFToStream(Stream, False);
    Stream.Position := 0;
    rvOutput.LoadRVFFromStream(Stream);
    Stream.Free;
    // processing the template
    FillFields(rvOutput.RVData, 0);
    rvOutput.Format;
  finally
    Screen.Cursor := crDefault;
  end;
end;

{==============================================================================}
{     The code below this line can be copied your application unchanged        }
{==============================================================================}
// Returns the field type by the field name.
// For field names containing ':', it assumes format <DATASET>:<DATAFIELD>,
// dataset is requested using GetDataSet().
// For field names without ':', it calls GetNonDBFieldType().
function TForm1.GetFieldType(const FieldName: String): TMMFieldType;
var DataSet: TDataSet;
    DataFieldDef: TFieldDef;
    p: Integer;
    DataSetName, DataField: String;
begin
  p := Pos(':', FieldName);
  if p=0 then
    Result := GetNonDBFieldType(FieldName)
  else begin
    Result := mmftError;
    DataSetName := Copy(FieldName, 1, p-1);
    DataField   := Copy(FieldName, p+1, Length(FieldName));
    DataSet := GetDataSet(DataSetName);
    if DataSet=nil then
      exit;
    DataFieldDef := DataSet.FieldDefList.Find(DataField);
    if DataFieldDef=nil then
      exit;
    if DataFieldDef.FieldClass=TGraphicField then
      Result := mmftPicture
    else
      Result := mmftText;
  end;
end;
{------------------------------------------------------------------------------}
// Returns the current text field value by the field name.
// For field names containing ':', it assumes format <DATASET>:<DATAFIELD>,
// dataset is requested using GetDataSet().
// For field names without ':', it calls GetNonDBFieldValue().
function TForm1.GetFieldValueFromDatabase(const FieldName: String): String;
var DataSet: TDataSet;
    p: Integer;
    DataSetName, DataField: String;
begin
  Result := '';
  p := Pos(':', FieldName);
  if p=0 then begin
    Result := GetNonDBFieldValue(FieldName);
    exit;
  end;
  DataSetName := Copy(FieldName, 1, p-1);
  DataField   := Copy(FieldName, p+1, Length(FieldName));
  DataSet := GetDataSet(DataSetName);
  if DataSet=nil then
    exit;
  Result := DataSet.FieldByName(DataField).AsString;
end;
{------------------------------------------------------------------------------}
// Returns the current graphic field value by the field name.
// For field names containing ':', it assumes format <DATASET>:<DATAFIELD>,
// dataset is requested using GetDataSet().
// Field names without ':' are not supported, but you can modify this method
// by adding GetNonDBPicture() and an analog of GetNonDBFieldValue().
function TForm1.GetPicture(const FieldName: String): TGraphic;
var DataSet: TDataSet;
    p: Integer;
    DataSetName, DataField: String;
begin
  Result := nil;
  p := Pos(':', FieldName);
  if p=0 then
    exit;
  DataSetName := Copy(FieldName, 1, p-1);
  DataField   := Copy(FieldName, p+1, Length(FieldName));
  DataSet := GetDataSet(DataSetName);
  if DataSet=nil then
    exit;
  Result := TBitmap.Create;
  try
    Result.Assign(DataSet.FieldByName(DataField))
  except
    Result.Assign(rvsOutput.InvalidPicture.Graphic);
  end;
end;
{------------------------------------------------------------------------------}
// Parses an item text for a potential detail table (TRVTableItemInfo)
// This text has form <DATASET> <FIRST REPEATED ROW> <COUNT OF REPEATED ROWS>.
// Is text is parsed correctly, returns True.
function TForm1.ParseTableText(s: String; RowCount: Integer;
  var DataSet: TDataSet;
  var RepRowStart, RepRowCount: Integer): Boolean;
var p: Integer;
begin
  Result := False;
  p := Pos(' ', s);
  if p=0 then
    exit;
  DataSet := GetDataSet(Copy(s, 1, p-1));
  if DataSet=nil then
    exit;
  s := Copy(s, p+1, Length(s));
  p := Pos(' ', s);
  if p=0 then
    exit;
  RepRowStart := StrToIntDef(Copy(s, 1, p-1), -1);
  if (RepRowStart<0) or (RepRowStart>=RowCount) then
    exit;
  s := Copy(s, p+1, Length(s));
  RepRowCount := StrToIntDef(Copy(s, 1, p-1), 0);
  if (RepRowCount<=0) or (RepRowStart+RepRowCount-1>=RowCount) then
    exit;
  Result := True;
end;
{------------------------------------------------------------------------------}
// This function iterates through all items in RVData starting from StartItemNo,
// searches for strings inside {}, calls InsertField()
// and replaces text between {} (including {}) with the text/picture/table.
procedure TForm1.FillFields(RVData: TCustomRVData; StartItemNo: Integer);
var i,j,r,c: Integer;
    table: TRVTableItemInfo;
    FieldName, OldFilter: String;
    s: String;
    Stream: TMemoryStream;
    FieldStart, RepRowStart, RepRowCount: Integer;
    DataSet: TDataSet;
begin
  i := StartItemNo;
  while i<RVData.ItemCount do begin
    if RVData.GetItemStyle(i)=rvsTable then begin
      // --- TABLES ---
      table := TRVTableItemInfo(RVData.GetItem(i));
      // Checking if the table is linked to a dataset
      if not  ParseTableText(RVData.GetItemText(i), table.RowCount,
        DataSet, RepRowStart, RepRowCount) then begin
        DataSet := nil;
        RepRowStart := -1;
        RepRowCount := 0;
      end;
      // Processing unlinked tables, or fixed rows of linked tables
      for r := 0 to table.Rows.Count-1 do
        if (r<RepRowStart) or (r>=RepRowStart+RepRowCount) then
          for c := 0 to table.ColCount-1 do
            if table.Cells[r,c]<>nil then
              FillFields(table.Cells[r,c].GetRVData, 0);
      if DataSet<>nil then begin // Processing a table linked to dataset
        DataSet.Filter := Filter; // filtering according to the parent master table
        DataSet.Filtered := Filter<>'';
        if DataSet.RecordCount=0 then begin
          // if no matched records, delete repeated rows
          if table.RowCount=RepRowCount then begin
            RVData.DeleteItems(i, 1);
            continue;
          end;
          table.DeleteRows(RepRowStart, RepRowCount, False)
          end
        else begin
          Stream := TMemoryStream.Create;
          try
            // saving the repeated rows to a stream
            table.SaveRowsToStream(Stream, RepRowStart, RepRowCount);
            // inserting (RecordCount-1) records ("...-1" is because one instance
            // or repeated rows is already in the table)
            table.InsertRows(RepRowStart+RepRowCount, RepRowCount*(DataSet.RecordCount-1), -1, False);
            j := 0;
            OldFilter := Filter;
            // repeating rows
            DataSet.First;
            while not DataSet.Eof do begin
              if j>0 then begin
                // copying repeated rows
                Stream.Position :=0;
                table.LoadFromStreamEx(Stream, RepRowStart+j*RepRowCount);
              end;
              // replacing fields in repeated rows
              Filter := GetFilterFromMaster(DataSet); {!FILTER!}
              for r := RepRowStart+j*RepRowCount to RepRowStart+(j+1)*RepRowCount-1 do
                for c := 0 to table.Rows[r].Count-1 do
                  if table.Cells[r,c]<>nil then
                    FillFields(table.Cells[r,c].GetRVData, 0);
              inc(j);
              DataSet.Next;
            end;
            Filter := OldFilter;
          finally
            Stream.Free;
          end;
        end;
      end;
      end
    else if RVData.GetItemStyle(i)>=0 then begin
      // --- TEXT ITEMS ---
      // (searching for fields)
      s := RVData.GetItemText(i);
      FieldStart := 0;
      j := 1;
      while j<=Length(s) do begin
        if s[j]='{' then // may be this is a beginning of field code
          FieldStart := j
        else if (s[j]='}') and (FieldStart>0) then begin // field code is found
          FieldName :=  Copy(s, FieldStart+1, j-FieldStart-1);
          // deleting field code
          Delete(s, FieldStart, j-FieldStart+1);
          RVData.SetItemText(i, s);
          // inserting field value
          j := FieldStart;
          InsertField(RVData, i, j, FieldName);
          if RVData.GetItemStyle(i)<0 then
            s := ''
          else
            s := RVData.GetItemText(i);
          FieldStart := 0;
          continue;
        end;
        inc(j);
      end;
    end;
    inc(i);
  end;
end;
{------------------------------------------------------------------------------}
// Converting text from String to internal representation
function ConvertStringToItemText(const Text: String;
  UnicodeItem: Boolean; CodePage: Cardinal): TRVRawByteString;
begin
  {$IFDEF RVUNICODESTR} // <-- declared in RV_Defs.inc
  if UnicodeItem then
    Result := RVU_GetRawUnicode(Text)
  else
    Result := TRVAnsiString(Text);
  {$ELSE}
  if UnicodeItem then
    Result := RVU_AnsiToUnicode(CodePage, Text)
  else
    Result := Text;
  {$ENDIF}
end;
{------------------------------------------------------------------------------}
//  This function inserts a field value in the position (RVData, ItemNo, Pos).
//  After inserting, (RVData, ItemNo, Pos) is a position after the inserted item.
procedure TForm1.InsertField(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const FieldName: String);
var Item: TCustomRVItemInfo;
    Graphic: TGraphic;
begin
  case GetFieldType(FieldName) of
    mmftError:
      InsertMultilineText(RVData, ItemNo, Pos, '{error}');
    mmftText:
      InsertMultilineText(RVData, ItemNo, Pos, GetFieldValueFromDatabase(FieldName));
    mmftPicture:
      begin
        Graphic := GetPicture(FieldName);
        if Graphic=nil then begin
          InsertMultilineText(RVData, ItemNo, Pos, '{error: invalid picture}');
          exit;
        end;
        Item := TRVGraphicItemInfo.CreateEx(RVData, Graphic, rvvaBaseline);
        InsertItem(RVData, ItemNo, Pos, Item);
      end;
    mmftTable:
      begin
        Item := GetTable(RVData, FieldName);
        if Item=nil then begin
          InsertMultilineText(RVData, ItemNo, Pos, '{error: invalid table}');
          exit;
        end;
        InsertItem(RVData, ItemNo, Pos, Item);
      end;
  end;
end;
{------------------------------------------------------------------------------}
// Inserts item (for example, picture or table) in the position (RVData, ItemNo, Pos).
//  After inserting, (RVData, ItemNo, Pos) is a position after the inserted item.
//  This method uses some undocumented RichView functions. Just copy it to your
//  application.
procedure TForm1.InsertItem(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const Item: TCustomRVItemInfo);
var TextItem: TRVTextItemInfo;
    s1, s2: String;
    sr: TRVRawByteString;
    StyleNo, ParaNo: Integer;
    Unicode: Boolean;
    CodePage: TRVCodePage;
    FullWidthItem: Boolean;
begin
  s1 := RVData.GetItemText(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  ParaNo := RVData.GetItemPara(ItemNo);
  Item.ParaNo := ParaNo;
  StyleNo := RVData.GetItemStyle(ItemNo);
  Unicode := rvioUnicode in RVData.GetItem(ItemNo).ItemOptions;
  CodePage := RVData.GetItemCodePage(ItemNo);
  FullWidthItem := Item.GetBoolValue(rvbpFullWidth);
  Item.ParaNo := ParaNo;
  if s1<>'' then begin
   RVData.SetItemText(ItemNo, s1);
   if not FullWidthItem then
     Item.SameAsPrev := True;
   inc(ItemNo);
   end
  else begin
    if not FullWidthItem then
      Item.BR := RVData.GetItem(ItemNo).BR;
    Item.PageBreakBefore := RVData.GetItem(ItemNo).PageBreakBefore;
    RVData.DeleteItems(ItemNo, 1);
  end;
  Item.Inserting(RVData, sr, False);
  RVData.Items.InsertObject(ItemNo, sr, Item);
  Item.Inserted(RVData, ItemNo);
  Pos := 1;

  if s2<>'' then begin
    TextItem := RichViewTextItemClass.Create(RVData);
    TextItem.StyleNo := StyleNo;
    TextItem.ParaNo := ParaNo;
    if not FullWidthItem then
      TextItem.SameAsPrev := True;
    if Unicode then
      TextItem.ItemOptions := TextItem.ItemOptions+[rvioUnicode];
    sr := ConvertStringToItemText(s2, Unicode, CodePage);
    TextItem.Inserting(RVData, sr, False);
    RVData.Items.InsertObject(ItemNo+1, sr, TextItem);
    TextItem.Inserted(RVData, ItemNo+1);
    end
  else if FullWidthItem and (RVData.ItemCount>ItemNo+1) then
    RVData.GetItem(ItemNo+1).SameAsPrev := False;
end;
{------------------------------------------------------------------------------}
// Inserts multiline text Value in the position (RVData, ItemNo, Pos).
// After inserting, (RVData, ItemNo, Pos) is a position after the inserted text.
//  This method uses some undocumented RichView functions. Just copy it to your
//  application.
procedure TForm1.InsertMultilineText(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const Value: String);
var item: TRVTextItemInfo;
    s1, s2: String;
    sr: TRVRawByteString;
    sl: TStringList;
    i, StyleNo, ParaNo: Integer;
    Unicode: Boolean;
    CodePage: TRVCodePage;
begin
  if Value='' then
    exit;
  s1 := RVData.GetItemText(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  sl := TStringList.Create;
  sl.Text := Value;
  { workaround - TStringList ignores the last cr lf }
  if (Length(Value)>1) and
     (Value[Length(Value)-1]=#13) and (Value[Length(Value)]=#10) then
    sl.Add('');
  { end of workaround }
  if sl.Count=1 then begin
    RVData.SetItemText(ItemNo, s1+Value+s2);
    sl.Free;
    inc(Pos, Length(Value)-1);
    exit;
  end;
  RVData.SetItemText(ItemNo, s1+sl[0]);
  StyleNo := RVData.GetItemStyle(ItemNo);
  ParaNo := RVData.GetItemPara(ItemNo);
  Unicode := rvioUnicode in RVData.GetItem(ItemNo).ItemOptions;
  CodePage := RVData.GetItemCodePage(ItemNo);
  for i := 1 to sl.Count-1 do begin
    item := RichViewTextItemClass.Create(RVData);
    s1 := sl[i];
    if i=sl.Count-1 then
      s1 := s1+s2;
    item.StyleNo := StyleNo;
    item.ParaNo := ParaNo;
    if Unicode then
      item.ItemOptions := item.ItemOptions+[rvioUnicode];
    sr := ConvertStringToItemText(s1, Unicode, CodePage);
    item.Inserting(RVData, sr, False);
    RVData.Items.InsertObject(ItemNo+i, sr, item);
    item.Inserted(RVData, ItemNo+i);
  end;
  ItemNo := ItemNo+sl.Count-1;
  Pos := Length(sl[sl.Count-1])+1;
  sl.Free;
end;



end.
